#!/usr/bin/env python
# XMarshaL2 - pickle/unpickle python objects to XML 
#
# version:      $Id: XMarshaL2.py 92 2010-05-26 11:43:05Z jraedler $
# author:       Joerg Raedler jr@j-raedler.de
#
# Should work with python, jython
#
# this code is based on:
#   Recipe 355487: Pickle Python Objects to a DOM tree and vice versa
#   (http://code.activestate.com/recipes/355487/)
# by Uwe Schmitt. Thanks to the original author and to activestate!
#

import sys, types
import xml.dom.minidom as dom

objMarker = '__xmlDump__'
attMarker = '__xmlDumpAttributes__'
simpleTypes = {'int': int, 'float': float, 'str': str, 'unicode': unicode}
stringTypes = ('str', 'unicode')
stringMaxAttrLength = 20

class PickleError(Exception): pass
class PicklingError(PickleError): pass
class UnpicklingError(PickleError): pass


def markClassDumpable(*cls):
    """mark a class as dumpable"""
    [setattr(c, objMarker, True) for c in cls]


def markAttributesDumpable(cls, *a):
    """mark attributes as dumpable"""
    setattr(cls, attMarker, a)


def dumpable(cls):
    """class decorator to mark a class as dumpable"""
    markClassDumpable(cls)
    return cls


class Pickler(object):

    def __init__(self, prettyXML=True):
        self.doc = dom.Document()
        self.prettyXML = prettyXML

    def dumps(self, obj, elementName='root'):
        node = self.pickle(obj, self.doc, elementName)
        if self.prettyXML:
            return node.toprettyxml()
        else:
            return node.toxml()

    def pickle(self, obj, parentNode, elementName='root'):
        node = parentNode.createElement(elementName)
        typeStr = self.getType(obj)
        node.attributes['type'] = typeStr
        if hasattr(obj, objMarker) and getattr(obj, objMarker):
            node = self.pickleObjectWithAttributes(obj, node, parentNode, elementName)
        elif typeStr in simpleTypes:
            if typeStr in stringTypes and len(obj) > stringMaxAttrLength: 
                node.appendChild(parentNode.createTextNode(str(obj)))
            else:
                node.attributes['value'] = str(obj)
        elif isinstance(obj, dict):
            self.pickleDictItems(obj, node, parentNode)
        elif isinstance(obj, list):
            self.pickleListItems(obj, node, parentNode)
        elif isinstance(obj, tuple):
            self.pickleListItems(obj, node, parentNode)
        else:
            node.appendChild(parentNode.createTextNode(repr(obj)))
        return node

    def pickleDictItems(self, obj, node, parentNode):
        for key, value in obj.items():
            tempnode = parentNode.createElement('item')
            tempnode.appendChild(self.pickle(key, parentNode, 'key'))
            tempnode.appendChild(self.pickle(value, parentNode, 'value'))
            node.appendChild(tempnode)

    def pickleListItems(self, obj, node, parentNode):
        for idx, obj in enumerate(obj):
            tempnode = self.pickle(obj, parentNode, 'item')
            tempnode.attributes['index'] = str(idx)
            node.appendChild(tempnode)

    def pickleObjectWithAttributes(self, obj, node, parentNode, elementName):
        if hasattr(obj, attMarker):
            attributesToPickle = getattr(obj, attMarker)
        else:
            attributesToPickle = [ name for name in dir(obj) if not name.startswith("__") ]

        for name in attributesToPickle: 
            a = getattr(obj, name)
            if not hasattr(a, '__call__'):
                node.appendChild(self.pickle(a, parentNode, name))
        return node

    def getType(self, obj):
        cls = obj.__class__
        cname = cls.__name__
        try:
            cmodule = cls.__module__
        except:
            cmodule = '__builtin__'
        if cmodule in ('__main__', '__builtin__'):
            return cname
        else:
            return '%s.%s' % (cmodule, cname)


class Unpickler(object):

    def __init__(self):
        pass

    def loads(self, s):
        m = dom.parseString(s)
        return self.unpickle(m.childNodes[0])

    def unpickle(self, node):
        typeName= node.attributes['type'].value
        if typeName in simpleTypes:
            if node.hasAttribute('value'):
                v = node.attributes['value'].value
            else:
                v = self.getText(node.childNodes)
            return simpleTypes[typeName](v)
        elif typeName=='tuple':
            return self.unpickleTuple(node)
        elif typeName=='list':
            return self.unpickleList(node)
        elif typeName=='dict':
            return self.unpickleDict(node)
        else:
            # FIXME: does this work in any case?
            cls = sys.modules['__main__'].__dict__[typeName]
            obj = object.__new__(cls)
            for name, element in self.getElementChilds(node):
                setattr(obj, name, self.unpickle(element))
            return obj

    def unpickleList(self, node):
        li = []
        for name, element in self.getElementChilds(node):
            if name != 'item':
                raise UnpicklingException()
            idx = int(element.attributes['index'].value)
            obj = self.unpickle(element)
            li.append((idx, obj))
        li.sort()
        return [ item[1] for item in li ]

    def unpickleTuple(self, node):
        return tuple(self.unpickleList(node))

    def unpickleDict(self, node):
        dd = dict()
        for name, element in self.getElementChilds(node):
            if name != 'item':
                raise UnpicklingException()
            childList = self.getElementChilds(element)
            if len(childList) != 2:
                raise XMLUnpicklingException()
            for name, element in childList:
                if name=='key':
                    key = self.unpickle(element)
                elif name=='value':
                    value = self.unpickle(element)
            dd[key]=value
        return dd

    def getElementChilds(self, node):
        return [ (no.tagName.lower(), no) for no in node.childNodes if no.nodeType != no.TEXT_NODE ]

    def getText(self, nodelist):
        rc = ""
        for node in nodelist:
            if node.nodeType == node.TEXT_NODE:
                rc = rc + node.data
        return rc.strip()


p = Pickler()
u = Unpickler()
dumps = p.dumps
loads = u.loads
del p
del u
